#ifndef _SPARSEARRAY_MATH_METHODS_H_
#define _SPARSEARRAY_MATH_METHODS_H_

#include <Rdefines.h>

SEXP C_Math_SVT(
	SEXP x_dim,
	SEXP x_type,
	SEXP x_SVT,
	SEXP op
);

SEXP C_Math2_SVT(
	SEXP x_dim,
	SEXP x_type,
	SEXP x_SVT,
	SEXP op,
	SEXP digits
);

#endif  /* _SPARSEARRAY_MATH_METHODS_H_ */

